#include <cstdio>

#include "snd.hh"

using namespace std;

#define CONTENTDEBUG 0
song ReadSND(const char *fn)
{
    song song;
    FILE *fp = fopen(fn, "rb");
    
    if(!fp) { perror(fn); return song; }
    
    unsigned char Buf[64];
    fread(Buf, 16, 1, fp);
    if(Buf[0]!='S' || Buf[1]!='N' || Buf[2]!='D' || Buf[3]!=26)
    {
        fclose(fp);
        fprintf(stderr, "%s: Not a SND file\n", fn);
        return song;
    }
    
    int version = Buf[4];
    
    song.channelcount = Buf[5];
    song.rate         = Buf[6];
    
    if(!song.channelcount || song.channelcount>=0x20 || !song.rate)
    {
        fclose(fp);
        fprintf(stderr, "%s: Improper settings in file\n", fn);
        return song;
    }
    
    for(;;)
    {
        int c = fgetc(fp);
        if(c == EOF)break;
#if CONTENTDEBUG
        fprintf(stderr, "%02X  ", c);
#endif
        if(c == 0xFB)
        {
            int n = fgetc(fp), sl = fgetc(fp), sh = fgetc(fp);
            song.dpcmcache[n].resize(sl + sh*256);
            fread(&song.dpcmcache[n][0], 1, song.dpcmcache[n].size(), fp);
            continue;
        }
        if(c == 0xFE)
        {
            int c = fgetc(fp), t = fgetc(fp);
            
#if CONTENTDEBUG
            fprintf(stderr, "\n");
#endif
            SNDChannel &tmp = song.GetChan(c);
            tmp.type = (SNDChannel::chantype)t;
            continue;
        }
        if(c == 0xFF)
        {
#if CONTENTDEBUG
            fprintf(stderr, "\n");
#endif
            song.AddAge(1);
            continue;
        }
        if(c == 0xFD)
        {
#if CONTENTDEBUG
            fprintf(stderr, "\n");
#endif
            song.AddAge(fgetc(fp));
            continue;
        }
        if(version == 3 &&
           ((c & 0xF0) == 0x20
         || (c & 0xF0) == 0x40
         || (c & 0xF0) == 0x80
         || (c & 0xF0) == 0xC0))
        {
            int duty = c>>4, vol=c&15;
            int c=fgetc(fp), wl=fgetc(fp), wh=fgetc(fp);
            
            SNDChannel &tmp = song.GetChan(c);
            tmp.wavelen = wl + wh*256;
            tmp.vol = vol;
            tmp.sample = duty;
            continue;
        }
        if((unsigned)c < song.channelcount)
        {
            int wl = fgetc(fp), wh = fgetc(fp), vol = fgetc(fp);
            
#if CONTENTDEBUG
            fprintf(stderr, "\b\b\b\b[%u,%u,%u]  ",c,vol,wl+wh*256);
#endif
            SNDChannel &tmp = song.GetChan(c);
            tmp.wavelen = wl + wh*256;
            
            if(tmp.type == SNDChannel::undefined)
            {
                int t = (c==0 || c==1) ? SNDChannel::sqr : (c==2 ? SNDChannel::tri : SNDChannel::rnd);
                if(t == SNDChannel::sqr)tmp.sample = 4;
                tmp.type = (SNDChannel::chantype)t;
            }

            if(version==1) { vol /= 16; tmp.wavelen = 110000 / (wl+wh*256); }
            if(version==1 && c>1) tmp.wavelen /= 2;
            
            if(tmp.type == SNDChannel::pcm)
            {
                tmp.sample = vol;
                tmp.vol = vol ? 6 : 0;
            }
            else
                tmp.vol = vol;
            continue;
        }
    }
    song.GetChan(0);
    song.DeleteSilence();
    fclose(fp);
    fprintf(stderr, "Got %u rows (at rate %u, %u seconds)\n",
        song.rownum, song.rate, (song.rownum + song.rate - 1) / song.rate);
    return song;
}

void song::CheckRow()
{
    unsigned oldrownum = rownum;
    rownum += chanage;
    if(chanptr.size() <= rownum*channelcount)
    {
        const unsigned increment = 1024;
        chanptr.resize(((rownum + increment) & ~(increment-1)) * channelcount);
    }
    
    for(unsigned tmp=0; tmp<chanage; ++tmp)
    {
        unsigned t2 = tmp+1;
        unsigned src  = oldrownum * channelcount;
        unsigned dest = src + t2 * channelcount;
        for(unsigned c=0; c<channelcount; ++c)
        {
            chanptr[dest+c] = chanptr[src+c];
            chanptr[dest+c].age = chanptr[src+c].age + t2;
        }
    }
    chanage=0;
}

void song::DeleteSilence()
{
    unsigned rows=0;
    while(rows < rownum)
    {
        for(unsigned c=0; c<channelcount; ++c)
        {
            const SNDChannel &ch = ChanPtr(rows, c);
            if(ch.vol || ch.wavelen)goto done;
        }
        ++rows;
    }
done:
    if(rows)
    {
        chanptr.erase(chanptr.begin(), chanptr.begin() + rows * channelcount);
        rownum -= rows;
        fprintf(stderr, "%u row(s) of silence deleted\n", rows);
    }
}
